@@ -6,12 +6,8 @@ module Program.Mighty.Resource (
   , unlimit
   ) where
 
-#if __GLASGOW_HASKELL__ < 709
-import Control.Applicative
-#endif
-import Control.Exception
-import Control.Monad
 import System.Posix
+import UnliftIO.Exception
 
 ----------------------------------------------------------------
 
@@ -24,12 +20,15 @@ amIrootUser = (== 0) <$> getRealUserID
 -- | Setting user and group.
 setGroupUser :: String -- ^ User
              -> String -- ^ Group
-             -> IO ()
+             -> IO Bool
 setGroupUser user group = do
     root <- amIrootUser
-    when root $ do
+    if root then do
         getGroupEntryForName group >>= setGroupID . groupID
         getUserEntryForName user >>= setUserID . userID
+        return True
+      else
+        return False
 
 ----------------------------------------------------------------
 
